package de.fahsel.dewis;

import java.util.Calendar;
import java.util.Date;

/**
 * <p>Die DewisFacade Klasse stellt die zentrale Schnittstelle dar, über die auf die
 * gesamte Funktionalität dieser API zugegriffen werden kann.</p>
 * 
 * <p>Es wird nur eine Instanz dieser Klasse benötigt, über die potentiell sämtliche
 * Daten der über die API des Schachbundes http://www.schachbund.de/api.html erreichbar
 * sind.</p>
 * 
 * <p>Die API ist so konzipiert, dass alle Daten erst dann geladen werden, wenn Sie benötigt
 * werden und dann bis zum Ablauf der angegebenen Cache Lebensdauer von standardmäßig 1440 
 * Minuten (= 24 Stunden) nicht erneut geladen werden, um die eigene Anwendung zu
 * beschleunigen und den Server des Schachbundes zu entlasten.</p>
 * 
 * @author Peter Fahsel
 */
public class DewisFacade
{
	private Integer cacheLatency;
	private Date    cacheExpiration;
	
	/**
	 * Konstruktor für eine neue DewisFacade mit der Default Cache Lebensdauer von 24 Stunden
	 */
	public DewisFacade()
	{
		setCacheLatency(1440);
	}
	
	/**
	 * Konstruktor für eine neue DewisFacade mit selbst festgelegter Cache Lebensdauer
	 *  
	 * @param cacheLatency Cache Lebensdauer in Minuten
	 */
	public DewisFacade(Integer cacheLatency)
	{
		setCacheLatency(cacheLatency);
	}

	/**
	 * Abfrage eines Vereins per zps
	 * 
	 * Liefert ein {@link Club}-Objekt für die angegebene zps zurück, welches benutzt werden kann,
	 * um Spieler des Vereins abzufragen {@link Club#getPlayers()}
	 * 
	 * @param zps Die zps des abzufragenden Vereins
	 * @return {@link Club}
	 */
	public Club getClub(String zps)
	{
		checkCache();
		return Club.getInstance(zps);
	}

	/**
	 * Abfrage eines Verbands per zps
	 * 
	 * Liefert ein {@link Association}-Objekt für die angegebene zps zurück, welches benutzt werden 
	 * kann, um Spieler des Verbands abzufragen {@link Association#getPlayers()}
	 * 
	 * @param zps Die zps des abzufragenden Verbands
	 * @return {@link Association}
	 */
	public Association getAssociation(String zps)
	{
		checkCache();
		return Association.getInstance(zps);
	}
	
	/**
	 * Abfrage eines Spielers per pkz
	 * 
	 * Liefert ein {@link Player}-Objekt für die angegebene pkz zurück, welches alle Stammdaten, 
	 * Vereins- und Verbandsmitgliedschaften sowie die Turnierteilnahmen zugreifbar macht.
	 * 
	 * Die Dokumentation der Player-Klasse enthält hierzu detailierte Informationen.
	 * 
	 * @param pkz Die pkz des Abzufragenden Spieler
	 * @return {@link Player}
	 */
	public Player getPlayer(Integer pkz)
	{
		checkCache();
		return Player.getInstance(pkz);
	}
	
	/**
	 * @return Cache Lebensdauer in Minuten
	 */
	public Integer getCacheLatency()
	{
		return cacheLatency;
	}

	/**
	 * Setzt die Cache Lebensdauer neu. Der Cache wird dabei verworfen
	 * 
	 * @param cacheLatency Cache Lebensdauer in Minuten
	 */
	public void setCacheLatency(Integer cacheLatency)
	{
		this.cacheLatency = cacheLatency;
		clearCache();
	}
	
	/**
	 * Verwirft den Cache.
	 */
	public void clearCache()
	{
		Club.clearInstances();
		Association.clearInstances();
		Player.clearInstances();
		
		Calendar cal = Calendar.getInstance();
		cal.add(Calendar.MINUTE, cacheLatency);
		cacheExpiration = cal.getTime();
	}
	
	private void checkCache()
	{
		Date now = new Date();
		if (now.after(cacheExpiration))
		{
			clearCache();
		}
	}
}
