package de.fahsel.dewis;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.HttpURLConnection;
import java.net.URL;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Diese Klasse repräsentiert einen Spieler in der DWZ Liste des deutschen Schachbundes
 * 
 * @author Peter Fahsel
 *
 */
public class Player
{
	private static final Map<Integer, Player> instances = new HashMap<>();
	
	private static final String PLAYERS_URL = "http://www.schachbund.de/php/dewis/spieler.php?pkz=$$pkz$$&format=csv";
	
	private static final String CSV_PLAYER_HEADLINE        = "id|nachname|vorname|titel|dwz|dwzindex|fideid|fideelo|fidetitel|fidenation";
	private static final String CSV_ORGANISATIONS_HEADLINE = "zpsver|organisation|rang|auswerter";
	private static final String CSV_CLUB_HEADLINE          = "zpsver|vereinsname|zpsmgl|status";
	private static final String CSV_TOURNAMENTS_HEADLINE   = "turniercode|turniername|dwzalt|dwzaltindex|punkte|partien|nichtgewertet|erwartungswert|gegner|koeffizient|dwzneu|dwzneuindex|leistung";
	
	private Integer id;
	private String  lastName;
	private String  firstName;
	private String  title;
	private Integer dwz;
	private Boolean dwzSet;
	private Integer dwzIndex;
	private Boolean dwzIndexSet;
	private String  lastTournamentCode;
	private Date    lastTournamentEnd;
	private Boolean lastTournamentEndSet;
	private Integer fideId;
	private Boolean fideIdSet;
	private Integer fideElo;
	private Boolean fideEloSet;
	private String  fideTitle;
	private String  fideNation;
	
	private Boolean allDetailsLoaded = false;
	
	private Map<String, AbstractMembership> memberships;
	
	private List<Tournament>   tournaments;
	
	protected Player(Integer id)
	{
		setId(id);
	}
	
	protected Player(String id)
	{
		setId(id);
	}
	
	protected static void clearInstances()
	{
		synchronized(instances)
		{
			instances.clear();
		}
	}
	
	protected static Player getInstance(String id)
	{
		return getInstance(Integer.parseInt(id));
	}
	
	protected static Player getInstance(Integer id)
	{
		synchronized(instances)
		{
			Player instance = instances.get(id);
			if (instance == null)
			{
				instance = new Player(id);
				instances.put(id, instance);
			}
			return instance;
		}
	}
	
	/**
	 * @return ID (pkz) des Spielers
	 */
	public Integer getId()
	{
		return id;
	}
	
	protected void setId(Integer id)
	{
		this.id = id;
	}
	
	protected void setId(String id)
	{
		try
		{
			this.id = Integer.parseInt(id);
		}
		catch (Exception e)
		{
			this.id = null;
		}
	}
	
	/**
	 * @return Nachname des Spielers
	 * @throws IOException
	 */
	public String getLastName() throws IOException
	{
		checkSet(lastName);
		return lastName;
	}
	
	protected void setLastName(String lastName)
	{
		this.lastName = lastName;
	}
	
	/**
	 * @return Vorname des Spielers
	 * @throws IOException
	 */
	public String getFirstName() throws IOException
	{
		checkSet(firstName);
		return firstName;
	}
	
	protected void setFirstName(String firstName)
	{
		this.firstName = firstName;
	}
	
	/**
	 * @return Titel des Spielers
	 * @throws IOException
	 */
	public String getTitle() throws IOException
	{
		checkSet(title);
		return title;
	}
	
	protected void setTitle(String title)
	{
		this.title = title;
	}
	
	/**
	 * @return DWZ des Spielers
	 * @throws IOException
	 */
	public Integer getDwz() throws IOException
	{
		checkSet(dwzSet);
		return dwz;
	}
	
	/**
	 * <p>Hilfsfunktion zur Sortierung nach DWZ. Liefert entweder die DWZ des Spielers 
	 * zurück oder 0, wenn der Spieler keine DWZ hat.</p>
	 * 
	 * <p>Es ist zu beachten, dass die sich daraus ergebene Sortierung nicht unbedingt
	 * mit dem Rang in der jeweiligen Organisation übereinstimmt. Um den Rang eines Spielers
	 * in einer bestimmten Organisation zu erhalten, muss die mit
	 * {@link Player#getAssociationMembershipByZps(String)} die Mitgliedschaft in der
	 * Organisation abgefragt werden und dann per {@link AssociationMembership#getRank()}
	 * der Rang ermittelt werden. Dieser Weg ist jedoch <b>erheblich</b> teurer, da die
	 * API hierfür die Spielerdetails vom Server des Schachbunds laden muss, falls
	 * das nicht schon früher geschehen ist.</p>
	 * 
	 * @return DWZ des Spielers oder 0
	 * @throws IOException
	 */
	public Integer getSort() throws IOException
	{
		return (getDwz() == null) ? 0 : getDwz();
	}
	
	protected void setDwz(Integer dwz)
	{
		this.dwzSet = true;
		this.dwz    = dwz;
	}
	
	protected void setDwz(String dwz)
	{
		this.dwzSet = true;
		try
		{
			this.dwz = Integer.parseInt(dwz);
		}
		catch (Exception e)
		{
			this.dwz = null;
		}
	}
	
	/**
	 * @return DWZ Index des Spielers
	 * @throws IOException
	 */
	public Integer getDwzIndex() throws IOException
	{
		checkSet(dwzIndexSet);
		return dwzIndex;
	}
	
	protected void setDwzIndex(Integer dwzIndex)
	{
		this.dwzIndexSet = true;
		this.dwzIndex    = dwzIndex;
	}
	
	protected void setDwzIndex(String dwzIndex)
	{
		this.dwzIndexSet = true;
		try
		{
			this.dwzIndex = Integer.parseInt(dwzIndex);
		}
		catch (Exception e)
		{
			this.dwzIndex = null;
		}
	}
	
	/**
	 * @return Code des letzten gespielten Turniers
	 * @throws IOException
	 */
	public String getLastTournamentCode() throws IOException
	{
		checkSet(lastTournamentCode);
		return lastTournamentCode;
	}
	
	protected void setLastTournamentCode(String lastTournamentCode)
	{
		this.lastTournamentCode = lastTournamentCode;
	}
	
	/**
	 * @return Enddatum des letzten gespielten Turniers (entspricht letzter Auswertung)
	 * @throws IOException
	 */
	public Date getLastTournamentEnd() throws IOException
	{
		checkSet(lastTournamentEndSet);
		return lastTournamentEnd;
	}
	
	protected void setLastTournamentEnd(Date lastTournamentEnd)
	{
		this.lastTournamentEndSet = true;
		this.lastTournamentEnd    = lastTournamentEnd;
	}
	
	protected void setLastTournamentEnd(String lastTournamentEnd)
	{
		this.lastTournamentEndSet = true;
		try
		{
			SimpleDateFormat formatter = new SimpleDateFormat("yyyy-MM-dd");
			this.lastTournamentEnd = formatter.parse(lastTournamentEnd);
		}
		catch (Exception e)
		{
			this.lastTournamentEnd = null;
		}
	}
	
	/**
	 * @return FIDE ID des Spielers
	 * @throws IOException
	 */
	public Integer getFideId() throws IOException
	{
		checkSet(fideIdSet);
		return fideId;
	}
	
	protected void setFideId(Integer fideId)
	{
		this.fideIdSet = true;
		this.fideId    = fideId;
	}
	
	protected void setFideId(String fideId)
	{
		this.fideIdSet = true;
		try
		{
			this.fideId = Integer.parseInt(fideId);
		}
		catch (Exception e)
		{
			this.fideId = null;
		}
	}
	
	/**
	 * @return FIDE Elo des Spielers
	 * @throws IOException
	 */
	public Integer getFideElo() throws IOException
	{
		checkSet(fideEloSet);
		return fideElo;
	}
	
	protected void setFideElo(Integer fideElo)
	{
		this.fideEloSet = true;
		this.fideElo    = fideElo;
	}
	
	protected void setFideElo(String fideElo)
	{
		this.fideEloSet = true;
		try
		{
			this.fideElo = Integer.parseInt(fideElo);
		}
		catch (Exception e)
		{
			this.fideElo = null;
		}
	}
	
	/**
	 * @return FIDE Titel des Spielers
	 * @throws IOException
	 */
	public String getFideTitle() throws IOException
	{
		checkSet(fideTitle);
		return fideTitle;
	}
	
	protected void setFideTitle(String fideTitle)
	{
		this.fideTitle = fideTitle;
	}
	
	/**
	 * @return FIDE Nation des Spielers
	 * @throws IOException
	 */
	public String getFideNation() throws IOException
	{
		checkSet(fideNation);
		return fideNation;
	}
	
	protected void setFideNation(String fideNation)
	{
		this.fideNation = fideNation;
	}

	/**
	 * @return Liste der {@link Tournament} des Spielers
	 * @throws IOException
	 */
	public List<Tournament> getTournaments() throws IOException
	{
		checkSet(tournaments);
		return tournaments;
	}
	
	/**
	 * Gibt eine Kollektion aller Mitgliedschaften eines Spielers zurück. Das sind sowohl
	 * Vereins- {@link ClubMembership} als Auch Verbandsmitgliedschaften {@link AssociationMembership}
	 * 
	 * @return Liste der Mitgliedschaften 
	 * @throws IOException
	 */
	public Collection<AbstractMembership> getAllMemberships() throws IOException
	{
		loadDetails();
		return memberships.values();
	}

	/**
	 * Gibt eine Kollektion aller Vereinsmitgliedschaften eines Spielers zurück.
	 * 
	 * @return Liste der {@link ClubMembership} des Spielers
	 * @throws IOException
	 */
	public Collection<ClubMembership> getClubMemberships() throws IOException
	{
		List<ClubMembership> clubMemberships = new ArrayList<>();
		
		for (AbstractMembership membership: getAllMemberships())
		{
			if (membership instanceof ClubMembership)
			{
				clubMemberships.add((ClubMembership) membership);
			}
		}
		return clubMemberships;
	}
	
	/**
	 * Gibt eine Kollektion aller Verbandsmitgliedschaften eines Spielers zurück.
	 * 
	 * @return Liste der {@link AssociationMembership} des Spielers
	 * @throws IOException
	 */
	public Collection<AssociationMembership> getAssociationMemberships() throws IOException
	{
		List<AssociationMembership> clubMemberships = new ArrayList<>();
		
		for (AbstractMembership membership: getAllMemberships())
		{
			if (membership instanceof AssociationMembership)
			{
				clubMemberships.add((AssociationMembership) membership);
			}
		}
		return clubMemberships;
	}
	
	/**
	 * Gibt eine Kollektion der passiven Vereinsmitgliedschaften eines Spielers zurück.
	 * 
	 * @return Liste der passiven {@link ClubMembership} des Spielers
	 * @throws IOException
	 */
	public List<ClubMembership> getPassiveClubMemberships() throws IOException
	{
		List<ClubMembership> passiveClubMemberships = new ArrayList<>();
		
		for (ClubMembership membership: getClubMemberships())
		{
			if (membership.getStatus().equals("P"))
			{
				passiveClubMemberships.add(membership);
			}
		}
		return passiveClubMemberships;
	}
	
	/**
	 * Gibt eine die Vereinsmitgliedschaft eines Spielers anhand der zps des Clubs zurück,
	 * falls der Spieler in diesem Club Mitglied ist, ansonsten wird null zurückgegeben.
	 * 
	 * @return {@link ClubMembership}
	 * @throws IOException
	 */
	public ClubMembership getClubMembershipByZps(String zps) throws IOException
	{
		for (ClubMembership membership: getClubMemberships())
		{
			if (membership.getOrganisation().getZps().equals(zps))
			{
				return membership;
			}
		}
		return null;
	}
	
	/**
	 * Gibt eine die Verbandsmitgliedschaft eines Spielers anhand der zps des Clubs zurück,
	 * falls der Spieler in diesem Club Mitglied ist, ansonsten wird null zurückgegeben.
	 * 
	 * @return {@link AssociationMembership}
	 * @throws IOException
	 */
	public AssociationMembership getAssociationMembershipByZps(String zps) throws IOException
	{
		for (AssociationMembership membership: getAssociationMemberships())
		{
			if (membership.getOrganisation().getZps().equals(zps))
			{
				return membership;
			}
		}
		return null;
	}
	
	/**
	 * Gibt die aktive Vereinsmitgliedschaft des Spielers zurück, falls der Spieler eine aktive
	 * Mitgliedschaft hat, sonst wird null zurückgegeben.
	 * 
	 * @return {@link ClubMembership}
	 * @throws IOException
	 */
	public ClubMembership getActiveClubMembership() throws IOException
	{
		for (ClubMembership membership: getClubMemberships())
		{
			if (membership.getStatus() == null || membership.getStatus().equals(""))
			{
				return membership;
			}
		}
		return null;
	}
	
	protected void addOrReplaceMembership(AbstractMembership membership)
	{
		String key = membership.getClass().getName() + membership.getOrganisation().getZps();
		if (memberships == null)
		{
			memberships = new HashMap<>();
		}
		memberships.put(key, membership);
	}
	
	/**
	 * Lädt die Details (DWZ Karteikarte) eines Spielers und ergänzt das Objekt
	 * durch die Werte, die nur dort verfügbar sind.
	 * 
	 * @throws IOException
	 */
	public void loadDetails() throws IOException
	{
		if (allDetailsLoaded)
		{
			return;
		}
		
		URL               url = new URL(PLAYERS_URL.replace("$$pkz$$", getId().toString()));
		HttpURLConnection con = (HttpURLConnection) url.openConnection();
		String            inputLine;
		String            importMode = "";
		
		tournaments = new ArrayList<>();
		
		try (BufferedReader r = new BufferedReader(new InputStreamReader(con.getInputStream()))) {
			while ((inputLine = r.readLine()) != null)
			{
				if (inputLine.startsWith(CSV_PLAYER_HEADLINE))
				{
					importMode = "player";
				}
				else if (inputLine.startsWith(CSV_ORGANISATIONS_HEADLINE))
				{
					importMode = "organisations";
				}
				else if (inputLine.startsWith(CSV_CLUB_HEADLINE))
				{
					importMode = "club";
				}
				else if (inputLine.startsWith(CSV_TOURNAMENTS_HEADLINE))
				{
					importMode = "tournaments";
				}
				else
				{
					importLine(inputLine, importMode);
				}
			}
			allDetailsLoaded = true;
		}
	}

	private void importLine(String inputLine, String importMode)
	{
		switch (importMode)
		{
			case "player":
				importPlayer(inputLine);
				break;
		
			case "organisations":
				importOrganisation(inputLine);
				break;
		
			case "club":
				importClub(inputLine);
				break;
			
			case "tournaments":
				importTournament(inputLine);
				break;
			}
	}

	private void importPlayer(String inputLine)
	{
		String[] playerSplit = inputLine.split("\\|", -1);
		
		if (playerSplit.length < 10)
		{
			return;
		}
		
		setId(playerSplit[0]);
		setLastName(playerSplit[1]);
		setFirstName(playerSplit[2]);
		setTitle(playerSplit[3]);
		setDwz(playerSplit[4]);
		setDwzIndex(playerSplit[5]);
		setFideId(playerSplit[6]);
		setFideElo(playerSplit[7]);
		setFideTitle(playerSplit[8]);
		setFideNation(playerSplit[9]);
	}

	private void importOrganisation(String inputLine)
	{
		String[] organisationSplit = inputLine.split("\\|", -1);
		
		if (organisationSplit.length < 4)
		{
			return;
		}
		
		Association association = Association.getInstance(organisationSplit[0]);
		
		association.setName(organisationSplit[1]);
		
		AssociationMembership membership = new AssociationMembership(association, organisationSplit[2], organisationSplit[3]);
		
		addOrReplaceMembership(membership);
	}

	private void importClub(String inputLine)
	{
		String[] clubSplit = inputLine.split("\\|", -1);
		
		if (clubSplit.length < 4)
		{
			return;
		}
		
		Club club = Club.getInstance(clubSplit[0]);
		club.setName(clubSplit[1]);
		
		ClubMembership membership = new ClubMembership(club, clubSplit[3], clubSplit[2]);
		addOrReplaceMembership(membership);
	}
	
	private void importTournament(String inputLine)
	{
		String[] tournamentSplit = inputLine.split("\\|", -1);
		
		if (tournamentSplit.length < 13)
		{
			return;
		}
		
		Tournament tournament = new Tournament();
		
		tournament.setCode(tournamentSplit[0]);
		tournament.setName(tournamentSplit[1]);
		tournament.setOldDwz(tournamentSplit[2]);
		tournament.setOldDwzIndex(tournamentSplit[3]);
		tournament.setPoints(tournamentSplit[4]);
		tournament.setMatches(tournamentSplit[5]);
		tournament.setUnrated(tournamentSplit[6]);
		tournament.setExperienceValue(tournamentSplit[7]);
		tournament.setOpponentDwz(tournamentSplit[8]);
		tournament.setCoefficient(tournamentSplit[9]);
		tournament.setNewDwz(tournamentSplit[10]);
		tournament.setNewDwzIndex(tournamentSplit[11]);
		tournament.setPerformance(tournamentSplit[12]);
		
		tournaments.add(tournament);
	}
	
	private void checkSet(Object check) throws IOException
	{
		if (check == null)
		{
			loadDetails();
		}
	}
}
